@@ -106,6 +106,15 @@ static const char str_prefix_serial_airspy[STR_PREFIX_SERIAL_AIRSPY_SIZE] =
 
 #define SERIAL_AIRSPY_EXPECTED_SIZE (26)
 
+#define GAIN_COUNT (22)
+
+uint8_t airspy_linearity_vga_gains[GAIN_COUNT] = { 13, 12, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 8, 7, 6, 5, 4 };
+uint8_t airspy_linearity_mixer_gains[GAIN_COUNT] = { 12, 12, 11, 9, 8, 7, 6, 6, 5, 0, 0, 1, 0, 0, 2, 2, 1, 1, 1, 1, 0, 0 };
+uint8_t airspy_linearity_lna_gains[GAIN_COUNT] = { 14, 14, 14, 13, 12, 10, 9, 9, 8, 9, 8, 6, 5, 3, 1, 0, 0, 0, 0, 0, 0, 0 };
+uint8_t airspy_sensitivity_vga_gains[GAIN_COUNT] = { 13, 12, 11, 10, 9, 8, 7, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4 };
+uint8_t airspy_sensitivity_mixer_gains[GAIN_COUNT] = { 12, 12, 12, 12, 11, 10, 10, 9, 9, 8, 7, 4, 4, 4, 3, 2, 2, 1, 0, 0, 0, 0 };
+uint8_t airspy_sensitivity_lna_gains[GAIN_COUNT] = { 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 12, 12, 9, 9, 8, 7, 6, 5, 3, 2, 1, 0 };
+
 static int cancel_transfers(airspy_device_t* device)
 {
 	uint32_t transfer_index;
@@ -624,6 +633,15 @@ static void airspy_open_device(airspy_device_t* device,
 
 						if (strncmp((const char*)serial_number, serial_number_expected, SERIAL_AIRSPY_EXPECTED_SIZE) == 0)
 						{
+#ifdef __linux__
+							/* Check whether a kernel driver is attached to interface #0. If so, we'll 
+							 * need to detach it.
+							 */
+							if (libusb_kernel_driver_active(dev_handle, 0))
+							{
+								libusb_detach_kernel_driver(dev_handle, 0);
+							}
+#endif
 							result = libusb_set_configuration(dev_handle, 1);
 							if (result != 0)
 							{
@@ -657,6 +675,15 @@ static void airspy_open_device(airspy_device_t* device,
 				if (libusb_open(dev, libusb_dev_handle) == 0)
 				{
 					dev_handle = *libusb_dev_handle;
+#ifdef __linux__
+					/* Check whether a kernel driver is attached to interface #0. If so, we'll 
+					 * need to detach it.
+					 */
+					if (libusb_kernel_driver_active(dev_handle, 0))
+					{
+						libusb_detach_kernel_driver(dev_handle, 0);
+					}
+#endif
 					result = libusb_set_configuration(dev_handle, 1);
 					if (result != 0)
 					{
@@ -905,6 +932,9 @@ extern "C"
 		int result;
 
 		libusb_clear_halt(device->usb_device, LIBUSB_ENDPOINT_IN | 1);
+
+		iqconverter_float_reset(device->cnv_f);
+		iqconverter_int16_reset(device->cnv_i);
 		
 		result = airspy_set_receiver_mode(device, RECEIVER_MODE_RX);
 		if( result == AIRSPY_SUCCESS )
@@ -1239,7 +1269,7 @@ extern "C"
 			if(length > 0)
 			{
 				memcpy(version, version_local, length-1);
-				version[length] = 0;
+				version[length-1] = 0;
 				return AIRSPY_SUCCESS;
 			}
 			else
@@ -1315,6 +1345,32 @@ extern "C"
 		}
 	}
 
+	int ADDCALL airspy_set_conversion_filter_float32(struct airspy_device* device, const float *kernel, const uint32_t len)
+	{
+		if (device->streaming)
+		{
+			return AIRSPY_ERROR_BUSY;
+		}
+
+		iqconverter_float_free(device->cnv_f);
+		device->cnv_f = iqconverter_float_create(kernel, len);
+
+		return AIRSPY_SUCCESS;
+	}
+
+	int ADDCALL airspy_set_conversion_filter_int16(struct airspy_device* device, const int16_t *kernel, const uint32_t len)
+	{
+		if (device->streaming)
+		{
+			return AIRSPY_ERROR_BUSY;
+		}
+
+		iqconverter_int16_free(device->cnv_i);
+		device->cnv_i = iqconverter_int16_create(kernel, len);
+
+		return AIRSPY_SUCCESS;
+	}
+
 	int ADDCALL airspy_set_lna_gain(airspy_device_t* device, uint8_t value)
 	{
 		int result;
@@ -1459,6 +1515,80 @@ extern "C"
 		}
 	}
 
+	int ADDCALL airspy_set_linearity_gain(struct airspy_device* device, uint8_t value)
+	{
+		int rc;
+		value = GAIN_COUNT - 1 - value;
+
+		if (value >= GAIN_COUNT)
+		{
+			value = GAIN_COUNT - 1;
+		}
+		else if (value < 0)
+		{
+			value = 0;
+		}
+
+		rc = airspy_set_mixer_agc(device, 0);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_lna_agc(device, 0);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_vga_gain(device, airspy_linearity_vga_gains[value]);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_mixer_gain(device, airspy_linearity_mixer_gains[value]);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_lna_gain(device, airspy_linearity_lna_gains[value]);
+		if (rc < 0)
+			return rc;
+
+		return AIRSPY_SUCCESS;
+	}
+
+	int ADDCALL airspy_set_sensitivity_gain(struct airspy_device* device, uint8_t value)
+	{
+		int rc;
+		value = GAIN_COUNT - 1 - value;
+
+		if (value >= GAIN_COUNT)
+		{
+			value = GAIN_COUNT - 1;
+		}
+		else if (value < 0)
+		{
+			value = 0;
+		}
+
+		rc = airspy_set_mixer_agc(device, 0);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_lna_agc(device, 0);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_vga_gain(device, airspy_sensitivity_vga_gains[value]);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_mixer_gain(device, airspy_sensitivity_mixer_gains[value]);
+		if (rc < 0)
+			return rc;
+
+		rc = airspy_set_lna_gain(device, airspy_sensitivity_lna_gains[value]);
+		if (rc < 0)
+			return rc;
+
+		return AIRSPY_SUCCESS;
+	}
+
 	int ADDCALL airspy_set_rf_bias(airspy_device_t* device, uint8_t value)
 	{
 		return airspy_gpio_write(device, GPIO_PORT1, GPIO_PIN13, value);
